#include "StdAfx.h"
#include "cthumbprovider.h"
#include "illa/codecmgr.h"
#include "illa/surfacemgr.h"
#include "orz/stream_windows.h"
#include "illa/swsurface.h"
#include "illa/config.h"
#include "illa/render.h"
#include "illa/filter.h"
#include "codecsetup.h"

using namespace Img;
using namespace Geom;

CPictusThumbnailProvider::CPictusThumbnailProvider():m_cRef(1) {
	CodecManagerSetup(&m_cfs);
}

CPictusThumbnailProvider::~CPictusThumbnailProvider() {}

IFACEMETHODIMP CPictusThumbnailProvider::QueryInterface(REFIID riid, __deref_out void **ppv) {
	static const QITAB qit[] =  {
		QITABENT(CPictusThumbnailProvider, IInitializeWithStream),
		QITABENT(CPictusThumbnailProvider, IThumbnailProvider),
		{ 0 },
	};
	return QISearch(this, qit, riid, ppv);
}

IFACEMETHODIMP_(ULONG) CPictusThumbnailProvider::AddRef() {
	return InterlockedIncrement(&m_cRef);
}

IFACEMETHODIMP_(ULONG) CPictusThumbnailProvider::Release() {
	ULONG cRef = InterlockedDecrement(&m_cRef);
	if (!cRef)
		delete this;

	return cRef;
}

IFACEMETHODIMP CPictusThumbnailProvider::Initialize(IStream *pStream, DWORD) {
	if(m_reader)
		return E_UNEXPECTED;
	if(!pStream)
		return S_FALSE;

	IStream* stream;
	HRESULT ret = pStream->QueryInterface(IID_PPV_ARGS(&stream));
	if(!SUCCEEDED(ret))
		return ret;

	IO::Stream::Ptr winStream(new IO::StreamWindows(stream));
	m_reader.reset(new IO::FileReader(winStream));

	return S_OK;
}

struct DimData {
	SizeInt sz;
	float scale;
};

DimData DetermineDimensions(UINT cx, Geom::SizeInt surfDims) {
	DimData d;
	Geom::SizeFloat factors = (float)cx / surfDims.StaticCast<float>();
	d.scale = Util::Min(factors.Width, factors.Height, 1.0f);
	d.sz = (surfDims * d.scale).StaticCast<int>();
	return d;
}

// IThumbnailProvider
IFACEMETHODIMP CPictusThumbnailProvider::GetThumbnail(UINT cx, HBITMAP *phbmp, WTS_ALPHATYPE *pdwAlpha) {
	try {
		if(!m_reader) return S_FALSE;

		Img::SurfaceFactory(new FactorySurfaceSoftware);
		Img::CodecFactoryStore cfs;
		CodecManagerSetup(&cfs);

		Img::Surface::Ptr s = LoadSurface(cx);
		if(!s) {
			OutputDebugString(TX("PictErr: Failed to load image."));
			return S_FALSE;
		}

		DimData outDims = DetermineDimensions(cx, s->GetSize());
		if(!IsPositive(s->GetSize()) || !IsPositive(outDims.sz)) {
			OutputDebugString(TX("PictErr: Image or thumbnail had non-positive image dimensions."));
			return S_FALSE;
		}
		OutputDebugString((TX("PictErr: Loaded image, dims:") + ToString(s->GetSize())).c_str());

		BITMAPINFO bmi = {};
		bmi.bmiHeader.biSize = sizeof(bmi.bmiHeader);
		bmi.bmiHeader.biWidth = outDims.sz.Width;
		bmi.bmiHeader.biHeight = -static_cast<LONG>(outDims.sz.Height);
		bmi.bmiHeader.biPlanes = 1;
		bmi.bmiHeader.biBitCount = 32;
		bmi.bmiHeader.biCompression = BI_RGB;

		Byte* pBits;
		*phbmp = CreateDIBSection(0, &bmi, DIB_RGB_COLORS, reinterpret_cast<void **>(&pBits), NULL, 0);
		if(!*phbmp) {
			OutputDebugString(TX("PictErr: Failed to create DIB section."));
			return S_FALSE;
		}
		Filter::FilterBuffer dst(outDims.sz, 4, pBits, outDims.sz.Width * 4);

		Img::FilterBufferAndLock src = GenerateFilterBuffer(s);

		Img::Properties props;
		props.ResampleFilter = Filter::FilterLanczos3;
		props.Zoom = outDims.scale; 
		props.RetainAlpha = true;

		Img::RenderToBuffer(dst, src.filterBuffer, s->GetFormat(), RectInt(PointInt(0, 0), s->GetSize()), props);
		if(HasAlpha(s->GetFormat()))
			Filter::Alpha::PremultiplyAlphaBuffer(dst);
		else
			Filter::Alpha::SetAlpha(dst, 0xff);

		*pdwAlpha = (HasAlpha(s->GetFormat()))?WTSAT_ARGB:WTSAT_RGB;

		return S_OK;
	}
	catch(Err::Exception& e) {
		OutputDebugString((TX("PictErr:") + e.Desc()).c_str());
	}
	catch(...) {}
	return S_FALSE;
}

AbstractCodec* CPictusThumbnailProvider::FindCodec() {
	const Img::CodecFactoryStore::InfoVector& iv = m_cfs.CodecInfo();

	for(size_t i = 0; i < iv.size(); ++i) {
		AbstractCodec* c = m_cfs.CreateCodec(i);
		m_reader->Seek(0, IO::SeekMethodBegin);
		if(!c->CanDetectFormat() || !c->LoadHeader(m_reader))
			delete c;
		else
			return c;
	}
	return 0;
}

Img::Surface::Ptr CPictusThumbnailProvider::LoadSurface(UINT cx) {
	AbstractCodec* c = FindCodec();
	if(c == 0) return Img::Surface::Ptr();
	if(c->Allocate(SizeInt(cx, cx)) == Img::AbstractCodec::AllocNotSupported)
		if(c->Allocate() != Img::AbstractCodec::AllocOk) return Img::Surface::Ptr();

	if(c->LoadImageData() == AbstractCodec::LoadFailed)
		return Img::Surface::Ptr();

	Img::ImageComposer::Ptr cmp = c->RequestImageComposer();
	return cmp->RequestCurrentSurface();
}


HRESULT CPictusThumbnailProvider_CreateInstance(REFIID riid, __deref_out void **ppv) {
	HRESULT hr;
	CPictusThumbnailProvider *pNew = new CPictusThumbnailProvider;
	if(pNew) {
		hr = pNew->QueryInterface(riid, ppv);
		pNew->Release();        
	}
	else
		hr = E_OUTOFMEMORY;

	return hr;
}
