#include "StdAfx.h"
#include "ic_imageloader.h"
#include "illa/image.h"
#include "illa/codecmgr.h"
#include "orz/fileops.h"
#include "orz/exception.h"

namespace Img {
	Geom::SizeInt ImageLoader::m_maxRes;
	boost::mutex ImageLoader::m_mxHints;

	IO::OpenErrorCode ImageLoader::OpenFile() {
		return IO::OpenSucceeded;
	}

	LoadState ImageLoader::GetState() {
		if(m_state == ILError) return LoadStateDone;
		if(m_state == ILUnprocessed) return LoadStateUnprocessed;
		if(m_state == ILLoadedHeader) return LoadStateHeader;
		if(m_state == ILLoadingFast) return LoadStateLoading;
		if(m_state == ILDoneFast) return LoadStateHeader;
		if(m_state == ILLoadingFull) return LoadStateLoading;
		if(m_state == ILDoneFull) return LoadStateDone;
		DO_THROW(Err::CriticalError, TX("Unknown internal state encountered: ") + ToString(m_state));
	}

	void ImageLoader::Deallocate() {
		image->Deallocate();
		Purge();
	}

	void ImageLoader::Abort() {
		if(codec == 0) return;
		codec->Terminate();
	}

	void ImageLoader::LoadHeader() {
		try {
			//reader.reset(new IO::FileReader(filename));
			if(!reader->Open()) {
				return FailState();
			}
			COND_STRICT(reader->IsOpen(), Err::FileNotOpen, TX("File should be open now."));

			codec = DetectAndLoadHeader(reader);
			if(codec == 0)
				return FailState();

			image->SetHeaderData(codec->GetSize(), codec->RequestImageComposer());
			m_state = ILLoadedHeader;
			m_sw.Stop();
		}
		catch(Err::Exception&) {
			FailState();
		}
	}

	void ImageLoader::Allocate() {
		COND_STRICT(codec, Err::InvalidCall, TX("Codec not created."));
		COND_STRICT(m_state == ILLoadedHeader || m_state == ILDoneFast, Err::InvalidCall, TX("Loader not in a valid state for this operation.")
			TX("Current state: ") + ToString(m_state));

		m_sw.Reset();
		m_sw.Start();
		Geom::SizeInt resHint = GetMaximumResolutionHint();

		if(m_state == ILLoadedHeader) {
			if(!IsPositive(resHint)) {
				if(m_maySkipFast)
					goto fullalloc;
				m_state = ILDoneFast;
			}
			else switch(codec->Allocate(resHint)) {
				case AbstractCodec::AllocNotSupported:
					if(m_maySkipFast)
						goto fullalloc;

					m_state = ILDoneFast;
					break;
				case AbstractCodec::AllocOk:
					m_state = ILLoadingFast;
					break;
				case AbstractCodec::AllocFailed:
					FailState();
					break;
			}

		}
		else {
fullalloc:
			switch(codec->Allocate()) {
				case AbstractCodec::AllocNotSupported:
					FailState();
					DO_THROW(Err::CriticalError, TX("Full allocation reported as unsupported."));
					break;
				case AbstractCodec::AllocOk:
					m_state = ILLoadingFull;
					break;
				case AbstractCodec::AllocFailed:
					FailState();
					break;
			}
		}
		m_sw.Stop();
	}

	AbstractCodec::LoadStatus ImageLoader::LoadImageData() {
		m_sw.Start();
		const AbstractCodec::LoadStatus s = codec->LoadImageData(m_state==ILLoadingFull);
		switch(s) {
			case AbstractCodec::LoadFinished:
			case AbstractCodec::LoadFailed:
				if(m_state == ILLoadingFast)
					m_state = ILDoneFast;
				if(m_state == ILLoadingFull)
					m_state = ILDoneFull;
				image->FinishImage(m_sw.Stop());
		}
		return s;
	}

	size_t ImageLoader::MemoryRequirements() {
		if(m_state == ILError) return 0;

		COND_WARNING(codec, Err::InvalidCall, TX("Codec not created."));
		COND_WARNING(m_state >= ILLoadedHeader, Err::InvalidCall, TX("Loader not in a valid state for this operation."));
		
		return codec->EstimateMemory();
	}

	ImageLoader::ImageLoader(CodecFactoryStore* cfs, Image* img, const String& fname)
		:m_cfs(cfs), image(img), codec(0), m_maySkipFast(false), m_state(ILUnprocessed),
		reader(new IO::FileReader(fname))/*, filename(fname)*/ {
		COND_STRICT(m_cfs, Err::InvalidParam, TX("CodecFactoryStore may not be null."));
	}
	ImageLoader::~ImageLoader() {
		Purge();
	}

	void ImageLoader::FailState() {
		Purge();
		image->FinishImage(m_sw.Stop());
		m_state = ILError;
	}

	void ImageLoader::Purge() {
		if(codec) delete codec;
		codec = 0;
		m_state = ILUnprocessed;
		reader->Close();
	}

	void ImageLoader::ResetAbort() {
		if(!codec) return;
		codec->ResetTerminate();
	}

	void ImageLoader::SetMaximumResolutionHint( const Geom::SizeInt& maxRes ) {
		boost::mutex::scoped_lock l(m_mxHints);
		m_maxRes = maxRes;
	}

	Geom::SizeInt ImageLoader::GetMaximumResolutionHint() {
		boost::mutex::scoped_lock l(m_mxHints);
		return m_maxRes;
	}

	String ImageLoader::Filename() const {
		if(!reader) return TX("");
		return reader->Name();
	}

	IO::FileReader::Ptr ImageLoader::GetFileReader() {
		return reader;
	}

	bool ImageLoader::FastPassFinished() {
		return (m_state >= ILDoneFast);
	}

	bool ImageLoader::Error() const {
		return m_state == ILError;
	}

	Img::AbstractCodec* ImageLoader::DetectAndLoadHeader(const IO::FileReader::Ptr reader) {
		COND_STRICT(m_cfs, Err::InvalidCall, TX("CodecFactoryStore was null."));
		Img::AbstractCodec* c = m_cfs->CreateCodec(IO::GetExtension(/*filename*/reader->Name()));
		if(c->LoadHeader(reader))
			return c;

		const CodecFactoryStore::InfoVector& iv = m_cfs->CodecInfo();
		for(size_t i = 0; i < iv.size(); ++i) {
			c = m_cfs->CreateCodec(i);
			reader->Seek(0, IO::SeekMethodBegin);
			if(!c->CanDetectFormat() || !c->LoadHeader(reader))
				delete c;
			else
				return c;
		}
		return 0;
	}

	LoadEvent ImageLoader::Load() {
		switch(GetState()) {
			case LoadStateUnprocessed:
				LoadHeader();
				return LoadEventHeaderLoaded;
			case LoadStateHeader:
				Allocate();
				return LoadEventAllocated;
			case LoadStateLoading:
				if (GetState() == LoadStateLoading &&  LoadImageData() == AbstractCodec::LoadFinished)
					return LoadEventDone;
		}
		return LoadEventNone;
	}

	void ImageLoader::AllowFastSkip( bool doAllow ) {
		m_maySkipFast = doAllow;
	}
}
