#include "StdAfx.h"
#include "w32_reg.h"

namespace Win {
	bool LogResult(LONG code) {
		bool success = (code == ERROR_SUCCESS);
		Log << (success?TX("Succeeded."):TX("Failed.")) << TX("\n");
		return success;
	}
	bool RegistryKey::Open(bool doCreate, bool readOnly) {
		Close();
		if (RegOpenKeyEx(m_root, m_name.c_str(), 0, KEY_READ | (readOnly?0:KEY_WRITE), &m_key) == ERROR_SUCCESS) return true;
		if (doCreate) {
			try {
				Create();
				return true;
			}
			catch(...) {
			}
		}
		return false;
	}

	void RegistryKey::Create() {
		Close();
		DWORD disp;
		COND_STRICT(RegCreateKeyEx(m_root, m_name.c_str(), 0, 0, REG_OPTION_NON_VOLATILE, KEY_READ|KEY_WRITE, 0, &m_key, &disp) == ERROR_SUCCESS,  Err::Win32RegistryError, TX("Couldn't create registry key."));
		Log << "Created key: " << m_name << "\n";
	}

	bool RegistryKey::Exist() {
		if (m_key != 0) return true;
		bool ret = Open(false);
		Close();
		return ret;
	}

	bool RegistryKey::ValueExist(const String& name) {
		COND_STRICT(m_key, Err::Win32RegistryError, TX("key was not open."));
		Log << TX("RegistryKey::ValueExist (") << m_name << TX("\\") << name << TX(") :");
		DWORD size = 0;
		DWORD type = 0;
		return LogResult(RegQueryValueEx(m_key, name.c_str(), 0, &type, 0, &size));
	}

	void RegistryKey::DeleteValue(const String& name) {
		COND_STRICT(m_key, Err::Win32RegistryError, TX("key was not open."));
		Log << TX("RegistryKey::DeleteValue(") << m_name << TX("\\") << name << TX(") :");
		LogResult(RegDeleteValue(m_key, name.c_str()));
	}

	void RegistryKey::Close() {
		if (m_key)
			RegCloseKey(m_key);

		m_key = 0;
	}

	std::vector<Byte> RegistryKey::BinaryValue(const String& name) {
		return GetValue(name, REG_BINARY);
	}

	String RegistryKey::StringValue(const String& name) {
		std::vector<Byte> data = GetValue(name, REG_SZ);
		return String(reinterpret_cast<Char*>(&data[0]));
	}

	std::vector<Byte> RegistryKey::GetValue( const String& name, DWORD reqtype ) {
		COND_STRICT(m_key, Err::Win32RegistryError, TX("key was not open."));
		Log << TX("RegistryKey::GetValue (") << m_name << TX("\\") << name << TX(") :");

		// Get string size
		DWORD size = 0;
		DWORD type = 0;

		if (RegQueryValueEx(m_key, name.c_str(), 0, &type, 0, &size) != ERROR_SUCCESS) {
			Log << TX("Failed, not found.\n");
			return std::vector<Byte>();
		}

		COND_STRICT(type == reqtype, Err::Win32RegistryError, TX("actual value was not of expected type."));

		std::vector<Byte> buf(size);
		if (RegQueryValueEx(m_key, name.c_str(), 0, &type, (LPBYTE)&buf[0], &size) == ERROR_SUCCESS) {
			Log << TX("Succeded.\n");
			return buf;
		}

		Log << TX("Failed, could not query data.\n");
		return std::vector<Byte>();
	}

	void RegistryKey::StringValue(const String& name, const String& value) {
		COND_STRICT(m_key, Err::Win32RegistryError, TX("key was not open."));
		Log << TX("RegistryKey::StringValue(") << m_name << TX("\\") << name << TX(", ") << value << TX(") :");
		LogResult(RegSetValueEx(m_key, name.c_str(), 0, REG_SZ, reinterpret_cast<const Byte*>(value.c_str()), static_cast<DWORD>(value.length() * sizeof(Char))));
	}

	DWord RegistryKey::DWORDValue(const String& name) {
		COND_STRICT(m_key, Err::Win32RegistryError, TX("key was not open."));
		DWORD size	= sizeof(DWord);
		DWORD type	= 0;
		DWORD data;

		Log << TX("RegistryKey::DWORDValue(") << m_name << TX("\\") << name << TX(") :");

		if(!LogResult(RegQueryValueEx(m_key, name.c_str(), 0, &type, (LPBYTE)&data, &size)))
			return 0;

		COND_STRICT(type == REG_DWORD, Err::Exception, TX("Actual value was not a DWORD."));
		return data;
	}

	void RegistryKey::DWORDValue(const String& name, const DWord& value) {
		COND_STRICT(m_key, Err::Win32RegistryError, TX("key was not open."));
		Log << TX("RegistryKey::DWORDValue(") << m_name << TX("\\") << name << TX(", ") << value << TX(") :");
		LogResult(RegSetValueEx(m_key, name.c_str(), 0, REG_DWORD, (BYTE*)(&value), sizeof(DWord)));
	}

	RegistryKey::RegistryKey(const Root root, const String& name):m_name(name), m_key(0), m_root(0) {
		m_root = DetermineRootHKEY(root);
	}

	RegistryKey::~RegistryKey() {
		Close();
	}


	bool RecursiveKeyDelete(const RegistryKey::Root root, const String& name) {
		HKEY hCurrentKey;
		HKEY hRoot = DetermineRootHKEY(root);
		RegOpenKeyEx(hRoot, name.c_str(), 0, KEY_READ, &hCurrentKey);

		TCHAR currChild[MAX_PATH];
		DWORD dwSize = MAX_PATH;
		HRESULT hr = RegEnumKeyEx(hCurrentKey, 0, currChild, &dwSize, 0, 0, 0, 0);
		while(hr == ERROR_SUCCESS) {
			String thisNode = name + TX("\\") + currChild;
			if(!RecursiveKeyDelete(root, thisNode)) break;

			dwSize = MAX_PATH;
			hr = RegEnumKeyEx(hCurrentKey, 0, currChild, &dwSize, 0, 0, 0, 0);
		}
		RegCloseKey(hCurrentKey);

		return (S_OK == RegDeleteKey(hRoot, name.c_str()));
	}

	HKEY DetermineRootHKEY(RegistryKey::Root r) {
		switch(r) {
			case RegistryKey::ClassesRoot:
				return HKEY_CLASSES_ROOT;
			case RegistryKey::CurrentUser:
				return HKEY_CURRENT_USER;
			case RegistryKey::LocalMachine:
				return HKEY_LOCAL_MACHINE;
			case RegistryKey::Users:
				return HKEY_USERS;
			case RegistryKey::CurrentConfig:
				return HKEY_CURRENT_CONFIG;
			default:
				DO_THROW(Err::InvalidParam, TX("root param was invalid."));
		}
	}
};
