#include "StdAfx.h"
#include "c_psp.h"
#include "orz/bitconvert.h"
#include "psp_chunkskipper.h"

const Byte	Img::CodecPSP::FileHeader[32]={0x50, 0x61, 0x69, 0x6E, 0x74, 0x20, 0x53, 0x68, 0x6F, 0x70, 0x20, 0x50, 0x72, 0x6F, 0x20, 0x49, 0x6D, 0x61, 0x67, 0x65, 0x20, 0x46, 0x69, 0x6C, 0x65, 0x0A, 0x1A, 0x00, 0x00, 0x00, 0x00};
const Byte	Img::CodecPSP::Gen_HeaderID[4]={0x7E, 0x42, 0x4B, 0x00};

namespace Img {
	using namespace Geom;
	using namespace Internal;

	bool CodecPSP::PerformLoadHeader(IO::FileReader::Ptr file, ImageInfo& info) {
		Byte header[32];
		if (file->Read(header, 1, 32) != 32) return false;

		for(int i = 0; i < 32; ++i)
			if (header[i] != FileHeader[i])
				return false;

		Word ver;
		file->ReadFull(&ver, 2);

		if (ver < PSPVersion5) return false;

		// Skip minor version
		file->Seek(2, IO::SeekMethodCurrent);

		// Read general image attributes
		Byte headid[4];
		if (file->Read(headid, 1, 4) != 4) return false;

		for(int i = 0; i < 4; ++i)
			if (headid[i] != Gen_HeaderID[i])
				return false;

		// Check block ID
		Word blockid;
		file->ReadFull(&blockid, 2);
		if (blockid != PSP_IMAGE_BLOCK) return false;

		DWord blen;
		file->ReadFull(&blen, 4);

		// Load image attributes
		file->Seek(4, IO::SeekMethodCurrent);
		long w, h;
		file->ReadFull(&w, 4);
		file->ReadFull(&h, 4);

		// Skip weird stuff
		file->Seek(8, IO::SeekMethodCurrent);
		Byte resmetric;
		file->ReadFull(&resmetric, 1);

		m_compression = PSP_COMP_NONE;
		file->ReadFull(&m_compression, 2);
		if (m_compression != PSP_COMP_NONE && m_compression != PSP_COMP_RLE)
			return false;

		m_bitDepth = 0;
		Word tmpBitdepth;
		file->ReadFull(&tmpBitdepth, 2);
		m_bitDepth = static_cast<Byte>(tmpBitdepth);

		if (m_bitDepth == 1 || m_bitDepth == 4 || m_bitDepth == 8)
			info.SurfaceFormat = Img::FormatIndex8;
		else if (m_bitDepth == 24)
			info.SurfaceFormat = Img::FormatXRGB8888;
		else
			return false;

		// Seek back to the first block
		file->Seek(36, IO::SeekMethodBegin);

		info.Dimensions = Geom::SizeInt(w, h);

		return true;
	}

	AbstractCodec::AllocationStatus CodecPSP::PerformAllocate() {
		GetSurface()->CreateSurface(GetSize(), GetFormat());
		GetSurface()->ClearSurface(Img::Color(0xff, 0, 0, 0));
		return AllocOk;
	}

	AbstractCodec::LoadStatus CodecPSP::PerformLoadImageData(IO::FileReader::Ptr file) {
		if (m_compression != PSP_COMP_NONE && m_compression != PSP_COMP_RLE)
			return LoadFailed;

		ParseFileStructure(file);

		while (DoTerminate() == false) {
			if (m_currScanline == GetSurface()->Height())
				return LoadFinished;

			int scansToProcess = Util::Min<int>(GetSurface()->Height() - m_currScanline, ScansPerChunk);
			Surface::LockedArea::Ptr area  = GetSurface()->LockSurface(RectInt(PointInt(0, m_currScanline), SizeInt(GetSurface()->Width(), scansToProcess)));
			Byte* currChunkStart = area->Buffer();

			if (DecodeScanlines(file, scansToProcess, GetSurface()->Width(), area->Stride(), GetSurface()->PixelSize(), currChunkStart) == false) {
				area->Unlock();
				return LoadFailed;
			}

			m_currScanline += scansToProcess;
			area->Unlock();
		}
		return LoadAborted;
	}

	CodecPSP::CodecPSP()
		:m_isStructureParsed(false),
		 m_currScanline(0)
	{}

	String readVariableLengthString(IO::FileReader::Ptr file) {
		Word length;
		file->ReadFull(&length, 2);
		String str;
		str.resize(length);

		for(int i = 0; i < length; ++i) {
			Byte c;
			file->ReadFull(&c, 1);
			str[i] = c;
		}
		return str;
	}

	Geom::RectInt readRect(IO::FileReader::Ptr file) {
		DWord coord[4];
		file->ReadFull(coord, 16);
		Geom::RectInt r(Geom::PointInt(coord[0], coord[1]), Geom::SizeInt(coord[2]-coord[0], coord[3]-coord[1]));
		return r;
	}

	bool CodecPSP::ReadBlockHeader( IO::FileReader::Ptr reader, BlockHeader* outHeader ) {
		Byte header_id[4];

		outHeader->PositionStart = reader->Position();

		reader->ReadFull(&header_id, 4);
		outHeader->ID = PSP_IMAGE_BLOCK;
		reader->ReadFull(&outHeader->ID, 2);
		outHeader->Length = 0;
		reader->ReadFull(&outHeader->Length, 4);

		outHeader->PositionEnd = outHeader->PositionStart + outHeader->Length + 10;

		for(int i = 0; i < 4; ++i)
			if (header_id[i] != Gen_HeaderID[i])
				return false;

		return true;
	}

	bool CodecPSP::ReadLayerBlock(IO::FileReader::Ptr reader, Layer* outLayer) {
		String layername(readVariableLengthString(reader));

		// Fetch layer type (PSPLayerType)
		outLayer->Type = LTUndefined;
		reader->ReadFull(&outLayer->Type, 1);

		readRect(reader); // silly data
		outLayer->SavedLayerArea = readRect(reader);

		reader->Seek(1, IO::SeekMethodCurrent);
		outLayer->BlendMode = LAYER_BLEND_NORMAL;
		reader->ReadFull(&outLayer->BlendMode, 1);
		reader->ReadFull(&outLayer->LayerFlags, 1);

		return true;
	}

	bool CodecPSP::ReadChannelBlock(IO::FileReader::Ptr reader, Channel* outChannel) {
		outChannel->CurrentSourcePosition = 0;

		reader->ReadFull(&outChannel->SizeCompressed, 4);
		reader->ReadFull(&outChannel->SizeUncompressed, 4);
		outChannel->TypeDIB = PSP_DIB_IMAGE;
		reader->ReadFull(&outChannel->TypeDIB, 2);
		outChannel->TypeChannel = PSP_CHANNEL_COMPOSITE;
		reader->ReadFull(&outChannel->TypeChannel, 2);

		return true;
	}

	void CodecPSP::ParseFileStructure(IO::FileReader::Ptr reader) {
		if (m_isStructureParsed == true)
			return;

		m_isStructureParsed = true;

		try {
			while(true) {
				BlockHeader currentBlockHeader;
				if(ReadBlockHeader(reader, &currentBlockHeader) == false)
					return;

				bool skip = true;

				switch(currentBlockHeader.ID) {
					case PSP_LAYER_START_BLOCK:
						skip = false;
						break;

					case PSP_TABLE_BANK_BLOCK:
						skip = false;
						{
							ChunkSkipper c(reader);

							Word numBitmaps = 0;
							reader->ReadFull(&numBitmaps, sizeof(numBitmaps));
							if (numBitmaps != 1)
								return;
						}
						break;

					case PSP_COLOR_BLOCK:
						{
							DWord numColors = 0;
							{
								ChunkSkipper c(reader);
								reader->ReadFull(&numColors, 4);
								if (numColors > 256)
									return;

							}

							Img::Palette pal;
							for (DWord i = 0; i < numColors; ++i) {
								DWord rgbQuad;
								reader->ReadFull(&rgbQuad, 4);
								rgbQuad |= 0xff000000;
								pal.Color(i, Color::FromDWord(rgbQuad));
							}
							GetSurface()->SetPalette(pal);
						}
						break;

					case PSP_LAYER_BLOCK:
						skip = false;
						{
							ChunkSkipper c(reader);
							ReadLayerBlock(reader, &m_layer);
						}
						{
							ChunkSkipper c(reader);
							Word numBitmaps;
							reader->ReadFull(&numBitmaps, 2);
							Word numChannels;
							reader->ReadFull(&numChannels, 2);
						}
						break;

					case PSP_CHANNEL_BLOCK:
						{
							Channel currentChannel;

							{
								ChunkSkipper c(reader);
								ReadChannelBlock(reader, &currentChannel);
							}

							currentChannel.PositionData = reader->Position();
							m_channels.push_back(currentChannel);
						}
						break;
				}

				// Skip the rest of the block
				if (skip)
					reader->Seek(currentBlockHeader.PositionEnd, IO::SeekMethodBegin);
			}
		}
		catch (Err::EndOfFile&) {
			return;
		}
	}

	bool CodecPSP::DecodeScanlines(IO::FileReader::Ptr reader, int scansToProcess, int width, size_t stride, size_t pixelSize, Byte* destination) {
		if (m_compression == PSP_COMP_NONE)
			return DecodeRawScanlines(reader, scansToProcess, width, stride, pixelSize, destination);
		else if (m_compression == PSP_COMP_RLE)
			return DecodeRLEScanlines(reader, scansToProcess, width, stride, pixelSize, destination);
		else
			return false;
	}

	bool CodecPSP::DecodeRawScanlines(IO::FileReader::Ptr reader, int scansToProcess, int width, size_t stride, size_t pixelSize, Byte* destination) {
		if(m_channels.size() == 1 && (m_bitDepth == 4 || m_bitDepth == 1)) {
			Channel& currChannel = m_channels.at(0);

			// Scanlines are padded on a four-byte boundary
			size_t bytesPerScanline = ((width + 7) / 8) * m_bitDepth;

			boost::scoped_array<Byte> inputBuffer(new Byte[bytesPerScanline]);
			reader->Seek(currChannel.PositionData + currChannel.CurrentSourcePosition, IO::SeekMethodBegin);

			for (int y = 0; y < scansToProcess; ++y) {
				reader->ReadFull(inputBuffer.get(), bytesPerScanline);

				Util::ConvertToBytes(destination, inputBuffer.get(), m_bitDepth, width);

				currChannel.CurrentSourcePosition += bytesPerScanline;
				destination += stride;
			}
		}
		else if (m_channels.size() == 1 && m_bitDepth == 8) {
			Channel& currChannel = m_channels.at(0);

			reader->Seek(currChannel.PositionData + currChannel.CurrentSourcePosition, IO::SeekMethodBegin);

			for (int y = 0; y < scansToProcess; ++y) {
				reader->ReadFull(destination, width);
				destination += stride;
			}
			currChannel.CurrentSourcePosition += scansToProcess * width;

		}
		else if (m_channels.size() == 3 && m_bitDepth == 24) {
			boost::scoped_array<Byte> inputBuffer(new Byte[width]);

			for (int channel = 0; channel < 3; ++channel) {
				Channel& currChannel = m_channels.at(2 - channel);
				reader->Seek(currChannel.PositionData + currChannel.CurrentSourcePosition, IO::SeekMethodBegin);

				Byte* currScanStart = destination + channel;

				for (int y = 0; y < scansToProcess; ++y) {
					Byte* currPixel = currScanStart;

					reader->ReadFull(inputBuffer.get(), width);

					for (int x = 0; x < width; ++x) {
						*currPixel = *(inputBuffer.get() + x);
						currPixel += pixelSize;
					}
					currScanStart += stride;
				}

				currChannel.CurrentSourcePosition += scansToProcess * width;
			}
		}
		else
			return false;

		return true;
	}

	bool CodecPSP::DecodeRLEScanlines( IO::FileReader::Ptr reader, int scansToProcess, int width, size_t stride, size_t pixelSize, Byte* destination ) {
		if (m_channels.size() == 1 && (m_bitDepth == 4 || m_bitDepth == 1)) {
			Channel& currChannel = m_channels.at(0);

			// Scanlines are padded on a four-byte boundary
			int bytesPerScanline = ((width + 7) / 8) * m_bitDepth;

			boost::scoped_array<Byte> inputBuffer(new Byte[bytesPerScanline]);
			reader->Seek(currChannel.PositionData + currChannel.CurrentSourcePosition, IO::SeekMethodBegin);
			m_fileStream.SetFileReader(reader);

			for (int y = 0; y < scansToProcess; ++y) {
				DecodeRLEChannel(reader, 1, bytesPerScanline, stride, 1, inputBuffer.get());
				Util::ConvertToBytes(destination, inputBuffer.get(), m_bitDepth, width);
				destination += stride;
			}

			currChannel.CurrentSourcePosition = m_fileStream.VirtualFilePosition() - currChannel.PositionData;
		}
		else if (m_channels.size() == 1 && m_bitDepth == 8) {
			Channel& currChannel = m_channels.at(0);

			reader->Seek(currChannel.PositionData + currChannel.CurrentSourcePosition, IO::SeekMethodBegin);
			m_fileStream.SetFileReader(reader);

			if (DecodeRLEChannel(reader, scansToProcess, width, stride, pixelSize, destination) == false)
				return false;

			currChannel.CurrentSourcePosition = m_fileStream.VirtualFilePosition() - currChannel.PositionData;
		}
		else if (m_channels.size() == 3 && m_bitDepth == 24) {
			for (size_t channel = 0; channel < 3; ++channel) {
				Channel& currChannel = m_channels.at(2 - channel);
				reader->Seek(currChannel.PositionData + currChannel.CurrentSourcePosition, IO::SeekMethodBegin);
				m_fileStream.SetFileReader(reader);
				Byte* currScanStart = destination + channel;
				if (DecodeRLEChannel(reader, scansToProcess, width, stride, pixelSize, currScanStart) == false)
					return false;

				currChannel.CurrentSourcePosition = m_fileStream.VirtualFilePosition() - currChannel.PositionData;
			}
		}
		else
			return false;

		return true;
	}

	bool CodecPSP::DecodeRLEChannel( IO::FileReader::Ptr reader, int scansToProcess, int width, size_t stride, size_t pixelSize, Byte* destination ) {
		for (int y = 0; y < scansToProcess; ++y) {
			int x = 0;
			while (x < width) {
				Byte count = m_fileStream.ReadByte();
				if (count > 128) {
					count -= 128;
					if (count + x > width)
						return false;

					Byte data = m_fileStream.ReadByte();
					for (int pixel = 0; pixel < count; ++pixel)
						*(destination + (x + pixel) * pixelSize) = data;

					x += count;
				}
				else {
					if (count + x > width)
						return false;

					for (int pixel = 0; pixel < count; ++pixel)
						*(destination + (x + pixel) * pixelSize) = m_fileStream.ReadByte();

					x += count;
				}
			}
			destination += stride;
		}
		return true;
	}
}
