#include "StdAfx.h"

#include "c_xyz.h"
#include "surface.h"
#include "surfacemgr.h"

namespace Img {
	using namespace Geom;

	CodecXYZ::CodecXYZ()
		:m_isInit(false),
		 m_x(0),
		 m_y(0),
		 m_paletteSet(false)
	{
		memset(&m_stream, 0, sizeof(m_stream));
	}

	bool CodecXYZ::PerformLoadHeader( IO::FileReader::Ptr file, ImageInfo& info ) {
		char header[4];
		if(file->Size() < 10) return false;
		file->Read(header, 1, 4);
		if(memcmp(header, "XYZ1", 4) != 0) return false;
		Word width, height;
		file->Read(&width, 2, 1);
		file->Read(&height, 2, 1);

		m_size = Geom::SizeInt(width, height);
		info.Dimensions = m_size;
		info.SurfaceFormat = Img::FormatIndex8;

		memset(&m_stream, 0, sizeof(m_stream));
		m_stream.next_in = m_inBuffer;

		inflateInit(&m_stream);
		m_isInit = true;

		return true;
	}

	AbstractCodec::AllocationStatus CodecXYZ::PerformAllocate() {
		GetSurface()->CreateSurface(m_size, Img::FormatIndex8);
		return AllocOk;
	}

	AbstractCodec::LoadStatus CodecXYZ::PerformLoadImageData( IO::FileReader::Ptr file ) {
		while(DoTerminate() == false) {
			FillBuffer(file);
			if (m_outData.size() == 0) {
				Cleanup();
				return LoadFailed;
			}

			if (m_paletteSet == false && m_outData.size() >= 768) {
				m_paletteSet = true;
				Img::Palette pal;
				for (int i = 0; i < 256; ++i) {
					pal.Color(i, Img::Color(0xff, m_outData[0], m_outData[1], m_outData[2]));
					m_outData.pop_front();
					m_outData.pop_front();
					m_outData.pop_front();
				}
				GetSurface()->SetPalette(pal);
			}

			if (m_paletteSet) {
				Img::Surface::LockedArea::Ptr area = GetSurface()->LockSurface(
					RectInt(
						PointInt(0, m_y),
						SizeInt(m_size.Width, m_size.Height - m_y)));

				int currY = 0;
				Byte* buf = area->Buffer();
				size_t stride = area->Stride();
				while (m_outData.size() > 0) {
					buf[currY * stride + m_x] = m_outData.front();
					m_outData.pop_front();
					m_x++;
					if (m_x == m_size.Width) {
						m_x = 0;
						m_y++;
						currY++;

						if (m_y == m_size.Height) {
							Cleanup();
							return LoadFinished;
						}
					}
				}
			}
		}
		return LoadAborted;
	}

	bool CodecXYZ::FillBuffer(IO::FileReader::Ptr file) {
		Byte in_buffer[InBufferSize];
		Byte out_buffer[OutBufferSize];

		size_t inRemaining = file->Read(in_buffer, 1, InBufferSize);
		m_stream.next_in = in_buffer;
		m_stream.avail_in = inRemaining;

		do  {
			m_stream.next_out = out_buffer;
			m_stream.avail_out = OutBufferSize;
			int ret = inflate(&m_stream, 0);
			if (ret != Z_OK && ret != Z_STREAM_END) {
				Cleanup();
				return false;
			}
			int num_decompressed = OutBufferSize - m_stream.avail_out;
			for (int i = 0; i < num_decompressed; ++i)
				m_outData.push_back(out_buffer[i]);
		} while (m_stream.avail_in);

		return true;
	}

	void CodecXYZ::Cleanup() {
		if (m_isInit) inflateEnd(&m_stream);
		m_isInit = false;
	}

	CodecXYZ::~CodecXYZ() {
		Cleanup();
	}
}
