#include "StdAfx.h"

#include "codec.h"
#include "surface.h"

namespace Img {
	bool AbstractCodec::LoadHeader(IO::FileReader::Ptr reader) {
		try {
			COND_STRICT(reader, Err::InvalidParam, TX("Reader was null."));
			m_fs	= reader->Size();
			m_file	= reader;

			m_info.SurfaceFormat = Img::FormatUndefined;
			m_info.Dimensions = Geom::SizeInt(-1, -1);

			if (PerformLoadHeader(reader, m_info)) {
				// Validate info. Invalid values are treated as codec bugs.
				COND_STRICT(m_info.SurfaceFormat != Img::FormatUndefined, Err::CodecError, TX("Surface format was set to invalid."));
				COND_STRICT((m_info.Dimensions.Width > 0) && (m_info.Dimensions.Height > 0), Err::CodecError, TX("Dimensions were set to an invalid value."));
				return true;
			}
		}
		catch(...) {}

//		m_file->Close();
		m_file.reset();
		return false;
	}

	AbstractCodec::AllocationStatus AbstractCodec::Allocate(const Geom::SizeInt& dimHint){
		if(dimHint == Geom::SizeInt(0, 0))
			return PerformAllocate();
		return PerformAllocate(dimHint);
	}

	size_t AbstractCodec::EstimateMemory() {
		return PerformEstimateMemory();
	}

	AbstractCodec::LoadStatus AbstractCodec::LoadImageData(bool mayReset) {
		COND_STRICT(m_file, Err::InvalidCall, TX("No file was set"));
		COND_STRICT(m_file->IsOpen(), Err::InvalidCall, TX("File not open"));

		LoadStatus l;
		try {
			l = PerformLoadImageData(m_file);
			m_isFinished = (l != LoadAborted);
		}
		catch(Err::Exception& e) {
			l = LoadFinished;
			m_isFinished = true;
		}

		if(m_isFinished && m_file && mayReset)
			m_file->Close();

		return l;
	}

	bool AbstractCodec::IsFinished() {
		return m_isFinished;
	}

	const Geom::SizeInt& AbstractCodec::GetSize() const {
		return m_info.Dimensions;
	}

	void AbstractCodec::Terminate() {
		m_doAbort = true;
	}

	AbstractCodec::AbstractCodec()
		:m_doAbort(false),
		 m_fs(0),
		 m_isFinished(false)
	{
		m_info.Dimensions = Geom::SizeInt(0, 0);
		m_info.SurfaceFormat = Img::FormatUndefined;
	}

	AbstractCodec::~AbstractCodec() {}

	bool AbstractCodec::DoTerminate() {
		return m_doAbort;
	}

	size_t AbstractCodec::PerformEstimateMemory() {
		return Area(GetSize()) * EstimatePixelSize(GetFormat());
	}

	Img::Format AbstractCodec::GetFormat() const {
		return m_info.SurfaceFormat;
	}

	void AbstractCodec::ResetTerminate() {
		m_doAbort = false;
	}

	AbstractCodec::AllocationStatus AbstractCodec::PerformAllocate() {
		AllocationStatus a = PerformAllocate(GetSize());
		if(a == AllocNotSupported) return AllocFailed;
		return a;
	}

	AbstractCodec::AllocationStatus AbstractCodec::PerformAllocate( const Geom::SizeInt& ) {
		return AllocNotSupported;
	}

	bool AbstractCodec::CanDetectFormat() const {
		return true;
	}
}
