#include "StdAfx.h"
#include "filter_bilinear.h"
#include "filter_int.h"

#define BSHIFT		11
#define BSHIFTx2	22
#define BMUL		2048

namespace Filter {
	namespace Scale {
		using namespace Geom;

		namespace Internal {
			void set_contrib(Contrib* contrib, DWord i, DWord max_i, DWord max_coord, DWord ofs) {
				DWord floor = ofs >> BSHIFT;

				if (i < max_i) {
					contrib[i].floor = floor;
					if ((contrib[i].floor + 1) >= max_coord) {
						contrib[i].floor = Util::Max<DWord>(2, max_coord) - 2;
						ofs = (contrib[i].floor << BSHIFT) + BMUL - 2;
					}
				}

				contrib[i].frac = ofs - (contrib[i].floor << BSHIFT);
				contrib[i].frac_inv = BMUL - contrib[i].frac;
			}
		}

		template <typename T, class U> inline void performBilinear(const FilterBuffer& source, FilterBuffer& dest, const Geom::RectInt& region, float zoom, U pixelConverter) {
			int width = region.Width();
			int height = region.Height();
			using namespace Internal;


			// Precalculate stuff
			boost::scoped_array<Contrib> x_contrib(new Contrib[width]);
			boost::scoped_array<Contrib> y_contrib(new Contrib[height]);

			for(int i = 0; i < width; ++i)
				set_contrib(x_contrib.get(), i, width, source.Dimensions.Width, static_cast<DWord>(((region.Left() + i + 0.25f) / zoom) * BMUL));

			for(int i = 0; i < height; ++i)
				set_contrib(y_contrib.get(), i, height, source.Dimensions.Height, static_cast<DWord>(((region.Top() + i + 0.25f) / zoom) * BMUL));

			Byte* destCurrentScanlinePtr = dest.BufferData;
			int rw = region.Width();

			for (int y = 0; y < region.Height(); ++y) {
				DWord floorY = y_contrib[y].floor;
				DWord iFracY = y_contrib[y].frac;
				DWord iFracYinv	= y_contrib[y].frac_inv;

				const T* currSourceScanUpper = reinterpret_cast<T*>(source.BufferData + floorY * source.Stride);
				const T* currSourceScanLower = reinterpret_cast<T*>(source.BufferData + (floorY + 1) * source.Stride);

				DWord* currDestPixel = reinterpret_cast<DWord*>(destCurrentScanlinePtr);

				for(int x = 0; x < rw; ++x) {
					DWord iFracX	= x_contrib[x].frac;
					DWord iFracXinv	= x_contrib[x].frac_inv;
					DWord fUL		= (iFracXinv * iFracYinv);
					DWord fUR		= (iFracX * iFracYinv);
					DWord fBL		= (iFracXinv * iFracY);
					DWord fBR		= (iFracX * iFracY);

					*(currDestPixel++) = pixelConverter.Assemble(
						currSourceScanUpper[x_contrib[x].floor], fUL,
						currSourceScanUpper[x_contrib[x].floor + 1], fUR,
						currSourceScanLower[x_contrib[x].floor], fBL,
						currSourceScanLower[x_contrib[x].floor + 1], fBR);
				}
				destCurrentScanlinePtr += dest.Stride;
			}
		}

		void Bilinear(const FilterBuffer& source, FilterBuffer& dest, const Geom::RectInt& region, Img::Format format, float zoom) {
			if (format == Img::FormatXRGB8888) {
				struct DWordToDWord {
					DWord Assemble(DWord ul, DWord fUL, DWord ur, DWord fUR, DWord bl, DWord fBL, DWord br, DWord fBR) const {
						return Img::ToARGBDWORD(
							0xff,
							(Img::ChannelARGB8888Red(ul) * fUL + Img::ChannelARGB8888Red(ur) * fUR + Img::ChannelARGB8888Red(bl) * fBL + Img::ChannelARGB8888Red(br) * fBR) >> BSHIFTx2,
							(Img::ChannelARGB8888Green(ul) * fUL + Img::ChannelARGB8888Green(ur) * fUR + Img::ChannelARGB8888Green(bl) * fBL + Img::ChannelARGB8888Green(br) * fBR) >> BSHIFTx2,
							(Img::ChannelARGB8888Blue(ul) * fUL + Img::ChannelARGB8888Blue(ur) * fUR + Img::ChannelARGB8888Blue(bl) * fBL + Img::ChannelARGB8888Blue(br) * fBR) >> BSHIFTx2);
					}
				};
				performBilinear<DWord>(source, dest, region, zoom, DWordToDWord());
			}
			else if (format == Img::FormatARGB8888) {
				struct DWordToDWord {
					DWord Assemble(DWord ul, DWord fUL, DWord ur, DWord fUR, DWord bl, DWord fBL, DWord br, DWord fBR) const {
						return Img::ToARGBDWORD(
							(Img::ChannelARGB8888Alpha(ul) * fUL + Img::ChannelARGB8888Alpha(ur) * fUR + Img::ChannelARGB8888Alpha(bl) * fBL + Img::ChannelARGB8888Alpha(br) * fBR) >> BSHIFTx2,
							(Img::ChannelARGB8888Red(ul) * fUL + Img::ChannelARGB8888Red(ur) * fUR + Img::ChannelARGB8888Red(bl) * fBL + Img::ChannelARGB8888Red(br) * fBR) >> BSHIFTx2,
							(Img::ChannelARGB8888Green(ul) * fUL + Img::ChannelARGB8888Green(ur) * fUR + Img::ChannelARGB8888Green(bl) * fBL + Img::ChannelARGB8888Green(br) * fBR) >> BSHIFTx2,
							(Img::ChannelARGB8888Blue(ul) * fUL + Img::ChannelARGB8888Blue(ur) * fUR + Img::ChannelARGB8888Blue(bl) * fBL + Img::ChannelARGB8888Blue(br) * fBR) >> BSHIFTx2);
					}
				};
				performBilinear<DWord>(source, dest, region, zoom, DWordToDWord());
			}
			else if (format == Img::FormatXRGB1555) {
				struct WordToDWord {
					DWord Assemble(Word ul, DWord fUL, Word ur, DWord fUR, Word bl, DWord fBL, Word br, DWord fBR) const {
						return Img::ToARGBDWORD(
							0xff,
							((Img::ChannelARGB1555Red(ul) * fUL + Img::ChannelARGB1555Red(ur) * fUR + Img::ChannelARGB1555Red(bl) * fBL + Img::ChannelARGB1555Red(br) * fBR) >> BSHIFTx2) << 3,
							((Img::ChannelARGB1555Green(ul) * fUL + Img::ChannelARGB1555Green(ur) * fUR + Img::ChannelARGB1555Green(bl) * fBL + Img::ChannelARGB1555Green(br) * fBR) >> BSHIFTx2) << 3,
							((Img::ChannelARGB1555Blue(ul) * fUL + Img::ChannelARGB1555Blue(ur) * fUR + Img::ChannelARGB1555Blue(bl) * fBL + Img::ChannelARGB1555Blue(br) * fBR) >> BSHIFTx2) << 3);
					}
				};
				performBilinear<Word>(source, dest, region, zoom, WordToDWord());
			}
			else if (format == Img::FormatARGB1555) {
				struct WordToDWord {
					DWord Assemble(Word ul, DWord fUL, Word ur, DWord fUR, Word bl, DWord fBL, Word br, DWord fBR) const {
						return Img::ToARGBDWORD(
							255 * ((Img::ChannelARGB1555Alpha(ul) * fUL + Img::ChannelARGB1555Alpha(ur) * fUR + Img::ChannelARGB1555Alpha(bl) * fBL + Img::ChannelARGB1555Alpha(br) * fBR) >> BSHIFTx2),
							((Img::ChannelARGB1555Red(ul) * fUL + Img::ChannelARGB1555Red(ur) * fUR + Img::ChannelARGB1555Red(bl) * fBL + Img::ChannelARGB1555Red(br) * fBR) >> BSHIFTx2) << 3,
							((Img::ChannelARGB1555Green(ul) * fUL + Img::ChannelARGB1555Green(ur) * fUR + Img::ChannelARGB1555Green(bl) * fBL + Img::ChannelARGB1555Green(br) * fBR) >> BSHIFTx2) << 3,
							((Img::ChannelARGB1555Blue(ul) * fUL + Img::ChannelARGB1555Blue(ur) * fUR + Img::ChannelARGB1555Blue(bl) * fBL + Img::ChannelARGB1555Blue(br) * fBR) >> BSHIFTx2) << 3);
					}
				};
				performBilinear<Word>(source, dest, region, zoom, WordToDWord());
			}
			else if (format == Img::FormatRGB565) {
				struct WordToDWord {
					DWord Assemble(Word ul, DWord fUL, Word ur, DWord fUR, Word bl, DWord fBL, Word br, DWord fBR) const {
						return Img::ToARGBDWORD(
							0xff,
							((Img::ChannelRGB565Red(ul) * fUL + Img::ChannelRGB565Red(ur) * fUR + Img::ChannelRGB565Red(bl) * fBL + Img::ChannelRGB565Red(br) * fBR) >> BSHIFTx2) << 3,
							((Img::ChannelRGB565Green(ul) * fUL + Img::ChannelRGB565Green(ur) * fUR + Img::ChannelRGB565Green(bl) * fBL + Img::ChannelRGB565Green(br) * fBR) >> BSHIFTx2) << 2,
							((Img::ChannelRGB565Blue(ul) * fUL + Img::ChannelRGB565Blue(ur) * fUR + Img::ChannelRGB565Blue(bl) * fBL + Img::ChannelRGB565Blue(br) * fBR) >> BSHIFTx2)) << 3;
					}
				};
				performBilinear<Word>(source, dest, region, zoom, WordToDWord());
			}
			else if (format == Img::FormatIndex8) {
				//COND_STRICT(source.Palette, Err::InvalidParam, TX("Source palette must be set for palettized surfaces."));

				struct PaletteToDWord {
					PaletteToDWord(const Img::Palette& p):m_p(p) {}

					DWord Assemble(Byte ul, DWord fUL, Byte ur, DWord fUR, Byte bl, DWord fBL, Byte br, DWord fBR) const {
						const Img::Color& cUL	= m_p.Color(ul);
						const Img::Color& cUR	= m_p.Color(ur);
						const Img::Color& cBL	= m_p.Color(bl);
						const Img::Color& cBR	= m_p.Color(br);

						DWord A = (cUL.A * fUL + cUR.A * fUR + cBL.A * fBL + cBR.A * fBR) >> BSHIFTx2;
						DWord R = (cUL.R * fUL + cUR.R * fUR + cBL.R * fBL + cBR.R * fBR) >> BSHIFTx2;
						DWord G = (cUL.G * fUL + cUR.G * fUR + cBL.G * fBL + cBR.G * fBR) >> BSHIFTx2;
						DWord B = (cUL.B * fUL + cUR.B * fUR + cBL.B * fBL + cBR.B * fBR) >> BSHIFTx2;

						return Img::ToARGBDWORD(A, R, G, B);
					}

					const Img::Palette& m_p;
				};
				performBilinear<Byte>(source, dest, region, zoom, PaletteToDWord(source.Palette));
			}
			else
				DO_THROW(Err::InvalidParam, TX("Format not supported:") + ToString(format));
		}
	}
}
