#include "StdAfx.h"

#include "types.h"

namespace Filter {
	namespace Scale {
		using namespace Geom;

		struct ConvertAXRGB8888 {
			DWord operator()(const DWord& v) const { return v; }
		};

		struct ConvertARGB1555 {
			DWord operator()(const Word& v) const { return Img::ConvertARGB1555ToARGB8888(v); }
		};

		struct ConvertXRGB1555 {
			DWord operator()(const Word& v) const { return Img::ConvertXRGB1555ToARGB8888(v); }
		};

		struct ConvertRGB565 {
			DWord operator()(const Word& v) const { return Img::ConvertRGB565ToARGB8888(v); }
		};

		struct ConvertIndexed {
			ConvertIndexed(const Img::Palette& p):m_p(p) {}
			DWord operator()(const Byte& v) const { return Img::ToARGBDWORD(m_p.Color(v)); }
			const Img::Palette& m_p;
		};

		template <typename T, class U, int SHIFT, int MUL>
		inline void performNearestResampleWithGivenPrecision(const FilterBuffer& source, FilterBuffer& dest, const Geom::RectInt& region, float zoom, U& pixelConverter) {
			int xSD = static_cast<int>(MUL / zoom);
			int xSstart	= xSD * region.Left();

			Byte* destCurrentScanlinePtr = dest.BufferData;
			int rh = region.Height();
			int rw = region.Width();

			for (int y = 0; y < rh; ++y) {
				const T* currSourceScan = reinterpret_cast<T*>(source.BufferData + (static_cast<size_t>((region.Top() + y) / zoom)) * source.Stride);
				DWord* currDestPixel = reinterpret_cast<DWord*>(destCurrentScanlinePtr);

				size_t xSofs = xSstart;

				for (int x = 0; x < rw; ++x) {
					*(currDestPixel++) = pixelConverter(currSourceScan[xSofs >> SHIFT]);
					xSofs += xSD;
				}
				destCurrentScanlinePtr += dest.Stride;
			}
		}

		template <typename T, class U>
		inline void performNearestResample(const FilterBuffer& source, FilterBuffer& dest, const Geom::RectInt& region, float zoom, U pixelConverter) {
			if (zoom > 10.0)
				performNearestResampleWithGivenPrecision<T, U, 18, 262144>(source, dest, region, zoom, pixelConverter);
			else
				performNearestResampleWithGivenPrecision<T, U, 11, 2048>(source, dest, region, zoom, pixelConverter);
		}

		void NearestNeighbor(const FilterBuffer& source, FilterBuffer& dest, const Geom::RectInt& region, Img::Format format, float zoom) {
			switch(format) {
				case Img::FormatARGB8888:
				case Img::FormatXRGB8888:
					performNearestResample<DWord>(source, dest, region, zoom, ConvertAXRGB8888());
					break;
				case Img::FormatARGB1555:
					performNearestResample<Word>(source, dest, region, zoom, ConvertARGB1555());
					break;
				case Img::FormatXRGB1555:
					performNearestResample<Word>(source, dest, region, zoom, ConvertXRGB1555());
					break;
				case Img::FormatRGB565:
					performNearestResample<Word>(source, dest, region, zoom, ConvertRGB565());
					break;
				case Img::FormatIndex8:
					performNearestResample<Byte>(source, dest, region, zoom, ConvertIndexed(source.Palette));
					break;
				default:
					DO_THROW(Err::InvalidParam, TX("Surface format not supported:") + ToString(format));
			}
		}
	}
}
