#include "StdAfx.h"
#include "surface.h"
#include "filter.h"
#include "config.h"

namespace Img {
	using Geom::PointInt;
	using Geom::SizeInt;
	using Geom::RectInt;

	bool Surface::CreateSurface(const SizeInt& size, Format format) {
		COND_STRICT(m_isCreated == false, Err::CriticalError, TX("Surface already created."));

		// Sanity check
		if ((size.Width <= 0) || (size.Height <= 0)) return false;
		if (size.Width > MaxSurfaceDim || size.Height > MaxSurfaceDim) return false;
		if (format >= FormatNum) return false;
		if (static_cast<size_t>(size.Width) * static_cast<size_t>(size.Height) * static_cast<size_t>(EstimatePixelSize(format)) > MaxSurfaceBytes)
			return false;

		m_swFormat	= format;
		m_size		= size;
		m_pixelSize	= EstimatePixelSize(m_swFormat);

		ForceDirty();

		bool toRet = onCreateSurface();

		if (toRet)
			m_isCreated = true;

		return toRet;
	}

	bool Surface::CopyRectToBuffer(Byte* pDest, int strideDest, const RectInt& sourceRect, const Img::Color& bgrColor, int brightness, int contrast, int gamma, bool applyAlpha) {
		onCopyAreaToBuffer(pDest, strideDest, sourceRect, bgrColor, brightness, contrast, gamma, applyAlpha);
		return true;
	}

	void Surface::ClearSurface(const Img::Color& col) {
		onClearSurface(col);
	}

	void Surface::CopySurface(Surface::Ptr source) {
		CopySurface(source, PointInt(0, 0));
	}

	void Surface::CopySurface(Surface::Ptr source, PointInt destinationTopLeft) {
		CopySurface(source, RectInt(PointInt(0, 0), source->GetSize()), destinationTopLeft);
	}

	void Surface::CopySurface(Surface::Ptr source, RectInt sourceAreaToCopy, PointInt destinationTopLeft) {
		onCopySurface(source, sourceAreaToCopy, destinationTopLeft);
	}

	void Surface::BlitSurfaceColorKey( Surface::Ptr source, PointInt destinationTopLeft, Byte colorKeyIndex ) {
		BlitSurfaceColorKey(source, RectInt(PointInt(0, 0), source->GetSize()), destinationTopLeft, colorKeyIndex);
	}

	void Surface::BlitSurfaceColorKey( Surface::Ptr source, RectInt sourceAreaToCopy, PointInt destinationTopLeft, Byte colorKeyIndex ) {
		COND_STRICT(source->GetFormat() == FormatIndex8, Err::InvalidParam, TX("Source buffer must be indexed."));
		onBlitSurfaceColorKey(source, sourceAreaToCopy, destinationTopLeft, colorKeyIndex);
	}

	Surface::LockedArea::Ptr Surface::LockSurface(const RectInt& region, LockMethod method) {
		m_strategyLocking->AquireLock(method);
		if(!RectInt(PointInt(0, 0), GetSize()).Contains(region))
			return LockedArea::Ptr();

		Byte* pData = onLockSurface(region, method);
		COND_STRICT(pData, Err::CriticalError, TX("Couldn't lock surface."));

		if (method == LockReadWrite) ForceDirty();

		return LockedArea::Ptr(new LockedArea(this, pData, region, method, onStride()));
	}

	Surface::LockedArea::Ptr Surface::LockSurface(LockMethod method) {
		return LockSurface(RectInt(PointInt(0,0), GetSize()), method);
	}

	void Surface::UnlockSurface(LockMethod method) {
		onUnlockSurface();

		m_strategyLocking->ReleaseLock(method);
	}

	void Surface::onClearSurface(const Img::Color& col) {
		FillRect(RectInt(PointInt(0, 0), GetSize()), col);
	}

	size_t Surface::onStride() const {
		return Width() * m_pixelSize;
	}

	Surface::Surface()
		:m_isDirty(true),
		 m_isCreated(false),
		 m_pixelSize(0),
		 m_swFormat(FormatUndefined)
	{}

	Surface::~Surface() {}

	bool Surface::IsDirty() {
		boost::mutex::scoped_lock l(m_mutexDirty);
		bool currDirty = m_isDirty;
		m_isDirty = false;
		return currDirty;
	}

	void Surface::SetPalette(const Img::Palette& pal) {
		m_swPal = pal;
		onPaletteUpdate();
	}

	Img::Palette Surface::GetPalette() {
		return m_swPal;
	}

	void Surface::ForceDirty() {
		boost::mutex::scoped_lock l(m_mutexDirty);
		m_isDirty = true;
	}

	const SizeInt& Surface::GetSize() const {
		return m_size;
	}

	int Surface::PixelSize() const {
		return m_pixelSize;
	}

	Img::Format Surface::GetFormat() const {
		return m_swFormat;
	}

	void Surface::SetLockStrategy(LockStrategy::Ptr strategy) {
		m_strategyLocking = strategy;
	}

	void Surface::onCopySurface(Surface::Ptr source, RectInt sourceAreaToCopy, PointInt destinationTopLeft) {
		if (source->GetFormat() == GetFormat())
			copyNoConversion(source, sourceAreaToCopy, destinationTopLeft);
		else if (GetFormat() == Img::FormatARGB8888 || GetFormat() == Img::FormatXRGB8888)
			copyConvertAnyTo32(source, sourceAreaToCopy, destinationTopLeft);
	}

	void Surface::copyNoConversion(Surface::Ptr source, RectInt &sourceAreaToCopy, PointInt destinationTopLeft) {
		LockedArea::Ptr lockedSource = source->LockSurface(sourceAreaToCopy, LockRead);
		Byte* pOldDest = lockedSource->Buffer();

		LockedArea::Ptr lockedDestination = LockSurface(RectInt(destinationTopLeft, sourceAreaToCopy.Dimensions()));
		Byte* pNewDest = lockedDestination->Buffer();

		int scans	= Util::Min(source->Height(), Height());
		int width	= Util::Min(source->Width(), Width());
		for(int i = 0; i < scans; ++i)
			memcpy(&pNewDest[i * lockedDestination->Stride()], &pOldDest[i * lockedSource->Stride()], width * PixelSize());
	}

	void Surface::copyConvertAnyTo32( Surface::Ptr source, RectInt &sourceAreaToCopy, PointInt destinationTopLeft ) {
		LockedArea::Ptr lockedDestination = LockSurface(RectInt(destinationTopLeft, sourceAreaToCopy.Dimensions()));
		Byte* pNewDest = lockedDestination->Buffer();

		source->CopyRectToBuffer(pNewDest, lockedDestination->Stride(), sourceAreaToCopy, Img::Color(0xff, 0xff, 0xff, 0xff), 0, 10, 10, false);
	}

	void Surface::onCopyAreaToBuffer(Byte* pDest, int strideDest, const RectInt& sourceRect, const Img::Color& bgrColor, int brightness, int contrast, int gamma, bool applyAlpha) {
		const SizeInt& dstSize(sourceRect.Dimensions());

		// Set up destination buffer
		LockedArea::Ptr lockedSource = LockSurface();

		Filter::FilterBuffer src(
			GetSize(),
			PixelSize(),
			lockedSource->Buffer(),
			lockedSource->Stride(),
			GetPalette());
		Filter::FilterBuffer dst(
			SizeInt(strideDest / 4, dstSize.Height),
			4,
			pDest,
			strideDest);
		Filter::Scale::AutoFilter(Filter::FilterDirectCopy, src, dst, sourceRect, GetFormat(), 1.0f);

		if (gamma != 10)
			Filter::Adjust::Gamma(dst, dst, RectInt(PointInt(0, 0), dst.Dimensions), PointInt(0,0), gamma);

		if ((brightness != 0) || (contrast != Img::ContrastStep))
			Filter::Adjust::BrightnessContrast(dst, dst, RectInt(PointInt(0, 0), dst.Dimensions), PointInt(0,0), brightness, contrast);

		if (HasAlpha(GetFormat()) && applyAlpha)
			Filter::Alpha::SolidColor(dst, dst, RectInt(PointInt(0, 0), dst.Dimensions), PointInt(0,0), bgrColor);
	}

	void Surface::onBlitSurfaceColorKey(Surface::Ptr source, RectInt sourceAreaToCopy, PointInt destinationTopLeft, Byte colorKeyIndex) {
		LockedArea::Ptr lockedSource = source->LockSurface();
		LockedArea::Ptr lockedDestination = LockSurface();

		Filter::FilterBuffer src(
			source->GetSize(),
			source->PixelSize(),
			lockedSource->Buffer(),
			lockedSource->Stride(),
			source->GetPalette());
		Filter::FilterBuffer dst(
			GetSize(),
			PixelSize(),
			lockedDestination->Buffer(),
			lockedDestination->Stride());
		Filter::Alpha::IndexedColorKey(src, dst, sourceAreaToCopy, destinationTopLeft, source->GetFormat(), colorKeyIndex);
	}

	void Surface::FillRect(const RectInt& areaToFill, const Img::Color& colorToSet) {
		onFillRect(areaToFill, colorToSet);
	}

	void Surface::onFillRect(const RectInt& areaToFill, const Color& colorToSet) {
		LockedArea::Ptr lockedDestination = LockSurface(areaToFill);
		Byte* pNewDest	= lockedDestination->Buffer();

		size_t scans	= areaToFill.Height();
		size_t width	= areaToFill.Width();
		size_t stride	= lockedDestination->Stride();

		if(PixelSize() == 1) {
			size_t rowsize	= width * PixelSize();
			for(size_t y = 0; y < scans; ++y) {
				memset(pNewDest, colorToSet.A, rowsize);
				pNewDest += stride;
			}
		}
		else if (Format() == FormatRGB565) {
			Word toSet = ToRGBWord565(colorToSet);

			for(size_t y = 0; y < scans; ++y) {
				Word* currentPixel = reinterpret_cast<Word*>(pNewDest);

				for(size_t x = 0; x < width; ++x) *(currentPixel++) = toSet;
				pNewDest += stride;
			}
		}
		else if((Format() == FormatXRGB1555) || (Format() == FormatARGB1555)) {
			Word	toSet = ToARGBWord1555(colorToSet);

			for(size_t y = 0; y < scans; ++y) {
				Word* currentPixel = reinterpret_cast<Word*>(pNewDest);
				for(size_t x = 0; x < width; ++x) *(currentPixel++) = toSet;
				pNewDest += stride;
			}
		}
		else if (PixelSize() == 4) {
			DWord	toSet = ToARGBDWORD(colorToSet);

			for(size_t y = 0; y < scans; ++y) {
				DWord* currentPixel = reinterpret_cast<DWord*>(pNewDest);

				for(size_t x = 0; x < width; ++x) *(currentPixel++) = toSet;
				pNewDest += stride;
			}
		}
	}

	void Surface::Deallocate() {
		onDeallocate();
		m_isCreated = false;
		m_size = Geom::SizeInt(0, 0);
		m_swFormat = FormatUndefined;
		m_pixelSize = 0;
	}

	size_t Surface::LockedArea::Stride() const {
		return m_stride;
	}

	FilterBufferAndLock GenerateFilterBuffer(Surface::Ptr src) {
		FilterBufferAndLock fbl;
		fbl.lock = src->LockSurface(Img::LockRead);
		fbl.filterBuffer.Construct(src->GetSize(), src->PixelSize(), fbl.lock->Buffer(), fbl.lock->Stride(), src->GetPalette());
		return fbl;
	}
}
