#ifndef ILLA_TYPES_H
#define ILLA_TYPES_H

namespace Err {
	struct CodecError:public Exception {
		CodecError(const String& msg):Exception(TX("Codec error: ") + msg) {}
	};

	struct InvalidHeader:public CodecError {
		InvalidHeader(const String& msg):CodecError(TX("InvalidHeader: ") + msg) {}
	};
}

namespace Img {
	class Palette;
	class Image;
	class Surface;

	enum Format {
		FormatUndefined,
		FormatARGB8888,
		FormatXRGB8888,
		FormatARGB1555,
		FormatXRGB1555,
		FormatRGB565,
		FormatIndex8,
		FormatNum
	};

	Byte EstimatePixelSize(Format sf);
	bool HasAlpha(Format sf);
};

#include "geom.h"

namespace Img {
	class Color {
	public:
		int A, R, G, B;

		Color(int cA, int cR, int cG, int cB)
			:A(cA), R(cR), G(cG), B(cB)
		{}

		static Color FromDWord(DWord dw) {
			return Color(static_cast<Byte>(dw >> 24), static_cast<Byte>((dw & 0xff0000) >> 16), static_cast<Byte>((dw & 0xff00) >> 8), static_cast<Byte>(dw&0xff));
		}

		Color() {}
	};

	inline std::basic_ostream<AChar>& operator<<(std::basic_ostream<AChar>& in, const Img::Color& c) {
		in << "(" << (int)c.A << ", " << (int)c.R << ", " << (int)c.G << ", " << (int)c.B << ")";
		return in;
	}
}

#include "color_conv.h"

namespace Img {
	static const float HueCap = 360.0f;
	static const float LumCap = 1.0f;
	static const float SatCap = 1.0f;

	struct HLSTriplet {
		float						H, L, S;
		HLSTriplet() {}
		HLSTriplet(const float _H, const float _L, const float _S)
			:H(_H), L(_L), S(_S)
		{}
	};

	HLSTriplet RGB_HLS(Byte r, Byte g, Byte b);
	Byte HueToRGB(float m1, float m2, float hue);

	Img::Color HLS_RGB(float h, float l, float s);

	static inline Img::Color HLS_RGB(HLSTriplet& hls) {
		return HLS_RGB(hls.H, hls.L, hls.S);
	}

	inline static Color HLS(float h, float l, float s) {
		return Img::Color(HLS_RGB(h, l, s));
	}

	inline bool operator==(const Color& lhs, const Color& rhs) {
		return ToARGBDWORD(lhs) == ToARGBDWORD(rhs);
	}

	inline bool operator!=(const Color& lhs, const Color& rhs) {
		return !(lhs == rhs);
	}

	class Palette {
	public:
		void Color(int index, const Color& col) {
			COND_WARNING(index < 256, Err::InvalidParam, TX("Color index out of bounds (> 255)"));
			COND_WARNING(index >= 0, Err::InvalidParam, TX("Color index out of bounds (< 0)"));

			m_colors[index] = col;
		}

		const Img::Color& Color(int index) const {
			COND_WARNING(index < 256, Err::InvalidParam, TX("Color index out of bounds (> 255)"));
			COND_WARNING(index >= 0, Err::InvalidParam, TX("Color index out of bounds (< 0)"));

			return m_colors[index];
		}

	private:
		Img::Color m_colors[256];
	};

	Img::Palette Grayscale(int minEntry = 0, int maxEntry = 255);

	bool operator==(const Palette& lhs, const Palette& rhs);
}


namespace Filter {
	enum FilterMode {
		FilterUndefined = 255,
		FilterDirectCopy = 0,
		FilterNearestNeighbor,
		FilterBilinear,
		FilterLanczos3,
		FilterNum
	};

	enum RotationAngle {
		RotateDefault,
		FlipX,
		FlipY,
		Rotate90,
		Rotate180,
		Rotate270,
		RotateUndefined
	};

	struct FilterBuffer {
		Geom::SizeInt Dimensions;
		int PixelSize;
		Byte* BufferData;
		size_t Stride;
		Img::Palette Palette;

		void Construct(const Geom::SizeInt& sz_, int ps_, Byte* data_, size_t stride_);
		void Construct(const Geom::SizeInt& sz_, int ps_, Byte* data_, size_t stride_, const Img::Palette& pal_);

		bool operator==(const FilterBuffer& rhs) const;
		bool operator!=(const FilterBuffer& rhs) const;

		FilterBuffer(const Geom::SizeInt& sz_ = Geom::SizeInt(0, 0), int ps_ = 0, Byte* data_ = 0, size_t stride_ = 0);
		FilterBuffer(const Geom::SizeInt& sz_, int ps_, Byte* data_, size_t stride_, const Img::Palette& pal);
	};
}

namespace Img {
	struct Properties {
		int						Brightness;
		int						Contrast;
		int						Gamma;
		Filter::FilterMode		ResampleFilter;
		bool					RetainAlpha;
		Img::Color				BackgroundColor;
		float					Zoom;
		Filter::RotationAngle	Angle;

		Properties();
	};
}

#endif
