#include "StdAfx.h"
#include "file.h"
#include "stream_file.h"

namespace IO {
	using boost::mutex;

	const String FileReader::Name() const {
		return m_stream->Name();
	}

	bool FileReader::Open() {
		return m_stream->Open();
	}

	void FileReader::Close() {
		m_stream->Close();
	}

	FileInt FileReader::Size() {
		return m_stream->Size();
	}

	Stream* FileReader::GetStream() {
		return m_stream.get();
	}

	size_t FileReader::Read(void* buf, size_t size, size_t items) {
		return m_stream->Read(buf, size, items);
	}

	void FileReader::ReadFull(void* buf, size_t size) {
		COND_WARNING(IsOpen(), Err::FileNotOpen, TX("File is not open."));

		if(m_stream->Read(buf, size, 1) !=1)
			DO_THROW(Err::EndOfFile, TX("Reached EOF in a call to ReadFull"));
	}

	void FileReader::Seek(FileInt position, SeekMethod m) {
		return m_stream->Seek(position, m);
	}

	FileInt FileReader::Position() const {
		return m_stream->Position();
	}

	bool FileReader::IsOpen() const {
		return m_stream->IsOpen();
	}

	FileReader::FileReader( const String& name )
		:m_stream(new StreamFile(name))
	{}

	FileReader::FileReader(Stream::Ptr s)
		:m_stream(s)
	{}

	bool FileWriter::Open(const String& name, bool append) {
		Close();   // Might be re-using the object.
		return (_wfopen_s(
			&m_file,
			name.c_str(),
			append?TX("ab"):TX("wb")) == 0);
	}

	void FileWriter::Close() {
		if (m_file != 0)
			fclose(m_file);

		m_file=0;
	}

	size_t FileWriter::Write(const void* buf, size_t size, size_t items) {
		COND_STRICT(m_file, Err::FileNotOpen, TX("File is not open."));

		return fwrite(buf, size, items, m_file);
	}

	FileWriter::FileWriter()
		:m_file(0)
	{}

	FileWriter::~FileWriter() {
		Close();
	}

	void replace_substrings(String& str, const String& toFind, const String& toWrite, size_t offset=0) {
		std::size_t pos = offset;
		while(pos != String::npos) {
			pos = str.find(toFind.c_str(), pos);
			if (pos != String::npos)
				str.replace(
					str.begin() + pos, str.begin() + pos + toFind.length(),
					toWrite.begin(), toWrite.end());
		}
	}

	String GetPath(const String& s) {
		String ret(s);
		replace_substrings(ret, TX("/"), TX("\\"));
		replace_substrings(ret, TX("\\\\"), TX("\\"), 1);

		std::size_t dot = ret.find_last_of(TX("/\\"));
		if (dot != String::npos)
			return ret.substr(0, dot) + TX("\\");

		return String(TX(""));
	}

	String GetFile(const String& s) {
		std::size_t dot = s.find_last_of(TX("/\\"));
		if (dot != String::npos)
			return s.substr(dot + 1, String::npos);

		return String(TX(""));
	}

	String GetTitle(const String& s) {
		std::size_t start	= s.find_last_of(TX("/\\"));
		std::size_t end		= s.find_last_of(TX('.'));

		if (start != String::npos)
			return s.substr(start + 1, end - start - 1);

		return s.substr(0, end);
	}

	String GetExtension(const String& s) {
		std::size_t dot = s.find_last_of(TX('.'));
		if (dot != String::npos) {
			// Make sure there were no / or \ afterwards
			if (s.find_first_of(TX("/\\"), dot) != String::npos) return String(TX(""));
			String ext = s.substr(dot + 1);
			return ext;
		}
		return String(TX(""));
	}
}
