#include "StdAfx.h"
#include "stream_file.h"
#include "fileops.h"

namespace IO {
	using boost::recursive_mutex;

#ifdef WIN32
	String StreamFile::Rename(const String& newFilename, HWND handle) {
		recursive_mutex::scoped_lock l(m_mutexAccess);
		bool wasOpen = IsOpen();
		FileInt prevPos = (wasOpen?performPosition():0);
		performClose();
		String ret = IO::Rename(m_name, newFilename, handle);
		if(ret != TX(""))
			m_name = ret;

		if(wasOpen) {
			performOpen();
			performSeek(prevPos, SeekMethodBegin);
		}
		return ret;
	}

	void StreamFile::Renamed( const String& newFilename ) {
		recursive_mutex::scoped_lock l(m_mutexAccess);
		m_name = newFilename;

		if(!IsOpen()) return;
		FileInt prevPos = performPosition();
		performClose();
		performOpen();
		performSeek(prevPos, SeekMethodBegin);
	}

	bool StreamFile::Delete(bool doRecycle, HWND handle) {
		recursive_mutex::scoped_lock l(m_mutexAccess);

		bool wasOpen = IsOpen();
		FileInt prevPos = (wasOpen?performPosition():0);
		performClose();

		if(doRecycle && IO::FileRecycle(m_name, handle))
			return true;
		if(!doRecycle && IO::FileDelete(m_name, handle))
			return true;

		if(wasOpen) {
			performOpen();
			performSeek(prevPos, SeekMethodBegin);
		}
		return false;
	}
#endif

	size_t StreamFile::performRead(void* buf, size_t size, size_t items) {
		recursive_mutex::scoped_lock l(m_mutexAccess);
		return fread(buf, size, items, m_file);
	}

	FileInt StreamFile::performPosition() const {
		recursive_mutex::scoped_lock l(m_mutexAccess);
		return _ftelli64(m_file);
	}

	void StreamFile::performClose() {
		recursive_mutex::scoped_lock l(m_mutexAccess);
		if(m_file)
			fclose(m_file);

		m_file = 0;
	}

	bool StreamFile::performOpen() {
		recursive_mutex::scoped_lock l(m_mutexAccess);
		m_size = 0;
		m_file = _wfsopen(m_name.c_str(), TX("rb"), _SH_DENYWR);

		if (m_file == 0) {
			switch(errno) {
			case EACCES:
				m_error = OpenAccessViolation;
				break;
			default:
				m_error = OpenFileNotFound;
				break;
			}
		}
		else {
			m_error = OpenSucceeded;

			// Precache file size
			performSeek(0, SeekMethodEnd);

			m_size = performPosition();
			performSeek(0, SeekMethodBegin);
		}
		return (m_file != 0);
	}

	void StreamFile::performSeek(FileInt position, SeekMethod m) {
		recursive_mutex::scoped_lock l(m_mutexAccess);
		unsigned long flag = 0;

		switch(m) {
			case SeekMethodBegin:
				flag = SEEK_SET;
				break;
			case SeekMethodCurrent:
				flag = SEEK_CUR;
				break;
			case SeekMethodEnd:
				flag = SEEK_END;
				break;
			default:
				DO_THROW(Err::InvalidParam, TX("Attempted to use an unsupported seek method."));
		}

		int ret = _fseeki64(m_file, position, flag);

		if (ret != 0)
			DO_THROW(Err::IOException, TX("Couldn't seek as requested."));
	}

	bool StreamFile::performIsOpen() const {
		recursive_mutex::scoped_lock l(m_mutexAccess);
		return m_file != 0;
	}

	FileInt StreamFile::performSize() {
		recursive_mutex::scoped_lock l(m_mutexAccess);
		return m_size;
	}

	String StreamFile::performName() const {
		return m_name;
	}

	StreamFile::StreamFile(const String& filename):m_name(filename), m_file(0), m_error(OpenSucceeded), m_size(0) {
	}

	StreamFile::~StreamFile() {
		performClose();
	}
}
