#include "StdAfx.h"
#include "orz/file.h"
#include "main.h"
#include "illa/codecmgr.h"
#include "illa/surface.h"
#include "orz/fileops.h"
#include "illa/image.h"
#include "illa/codec.h"

SUITE(CodecTests) {
	Img::CodecFactoryStore g_cfs;

	std::vector<String> ParseCsv(String filename) {
		std::vector<String> strings;
		IO::FileReader r(filename);
		r.Open();
		String curr;
		
		AChar c;
		while(r.Read(&c, sizeof(c), 1)) {
			if(c == TX(',') || c == 10 || c == 13) {
				if(curr.empty() == false)
					strings.push_back(curr);

				curr.clear();
			}
			else if (c != TX(' '))
				curr.push_back(c);

			if(curr == TX("//")) {
				while(c != 10 && c != 13 && r.Read(&c, sizeof(c), 1));
				curr.clear();
			}
		}
		if(curr.empty() == false)
			strings.push_back(curr);
		return strings;
	}

	enum Error {
		ErrFileNotFound,
		ErrLoadFailed,
		ErrSizeDiffer,
		ErrDataDiffer,
		ErrNoError
	};

	struct Loaded {
		Img::Surface::Ptr ptr;
		Error err;
	};

	Loaded load(Img::AbstractCodec* pCodec, const String filename) {
		Loaded l = {Img::Surface::Ptr(), ErrNoError };

		IO::FileReader::Ptr f(new IO::FileReader(filename));
		//f->Filename(filename);
		f->Open();
		if(f->IsOpen() == false) {
			l.err = ErrFileNotFound;
			return l;
		}
		l.err = ErrLoadFailed;
		if(pCodec->LoadHeader(f) == false) return l;
		if(pCodec->Allocate() != Img::AbstractCodec::AllocOk) return l;
		pCodec->LoadImageData();
		l.err = ErrNoError;

		l.ptr = pCodec->RequestImageComposer()->RequestCurrentSurface();
		return l;
	}

	bool isAlphaIsh(Img::Surface::Ptr p) {
		return Img::HasAlpha(p->GetFormat());
	}

	bool isIndexed(Img::Surface::Ptr p) {
		return p->GetFormat() == Img::FormatIndex8;
	}


	void clearTransparent(DWord* bufDWord, int dwordsToCheck) {
		for(int i = 0; i < dwordsToCheck; ++i)
			if(Img::ChannelARGB8888Alpha(bufDWord[i]) == 0)
				bufDWord[i] = 0;
	}


	Error CompareImages(String f1, String f2) {
		Loaded l1 = load(g_cfs.CreateCodec(IO::GetExtension(f1)), f1);
		if(l1.err != ErrNoError) return l1.err;
		Loaded l2 = load(g_cfs.CreateCodec(IO::GetExtension(f2)), f2);
		if(l2.err != ErrNoError) return l2.err;

		if(l1.ptr->GetSize() != l2.ptr->GetSize()) return ErrSizeDiffer;

		bool isOneAlpha = isAlphaIsh(l1.ptr) || isAlphaIsh(l2.ptr);
		bool isOneNotAlpha = !isAlphaIsh(l1.ptr) || !isAlphaIsh(l2.ptr);
		if(!isIndexed(l1.ptr) && !isIndexed(l2.ptr) && isOneAlpha == isOneNotAlpha) return ErrDataDiffer;

		// Convert to ARGB32
		boost::scoped_array<Byte> d1(new Byte[l1.ptr->Width() * l1.ptr->Height() * 4]);
		boost::scoped_array<Byte> d2(new Byte[l1.ptr->Width() * l1.ptr->Height() * 4]);

		l1.ptr->CopyRectToBuffer(d1.get(), l1.ptr->Width() * 4, Geom::RectInt(Geom::PointInt(0, 0), l1.ptr->GetSize()), Img::Color(0xff, 0, 0, 0), 0, 10, 10, false);
		l2.ptr->CopyRectToBuffer(d2.get(), l1.ptr->Width() * 4, Geom::RectInt(Geom::PointInt(0, 0), l1.ptr->GetSize()), Img::Color(0xff, 0, 0, 0), 0, 10, 10, false);

		// Clear any fully transparent pixels
		int dwordsToCheck = Geom::Area(l1.ptr->GetSize());
		clearTransparent(reinterpret_cast<DWord*>(d1.get()), dwordsToCheck);
		clearTransparent(reinterpret_cast<DWord*>(d2.get()), dwordsToCheck);

		// Compare data
		int bytesToCompare = Geom::Area(l1.ptr->GetSize()) * 4;
		Byte* bufRaw1 = d1.get();
		Byte* bufRaw2 = d2.get();
		for(int i = 0; i < bytesToCompare; ++i)
			if(bufRaw1[i] != bufRaw2[i])
				return ErrDataDiffer;

		return ErrNoError;
	}

	bool ShouldFail(const String& file) {
		try {
			Loaded l1 = load(g_cfs.CreateCodec(IO::GetExtension(file)), file);
			if(l1.err == ErrFileNotFound) return false;
			if(l1.err == ErrNoError) return false;
		}
		catch(Err::Exception) { }
		return true;
	}


	TEST(TestLoadCompare) {
		g_cfs.AddBuiltinCodecs();

		std::vector<String> files = ParseCsv(g_datapath + TX("\\loadcompare.txt"));
		for(std::vector<String>::iterator i = files.begin(); i != files.end(); ++i) {
			String a = *i;
			try {
				if(a[0] == TX('!')) {
					a = a.substr(1, String::npos);
					if(ShouldFail(g_datapath + TX("\\Codecs\\") + a) == false) {
						String wmsg = TX("Didn't report failure as expected: ") + a;
						AString msg(wmsg.begin(), wmsg.end());
						DECORATIVE_FAIL(msg.c_str());
					}
					continue;
				}
				i++;

				if(i == files.end())
					CHECK(false);
				String b = *i;
				switch(CompareImages(g_datapath + TX("\\Codecs\\") + a, g_datapath + TX("\\Codecs\\") + b)) {
					case ErrNoError:
						break;
					case ErrFileNotFound:
						{
							DECORATIVE_FAIL(ToAString(TX("File not found: ") + a + TX(" <=> ") + b).c_str());
							break;
						}
					case ErrSizeDiffer:
						{
							DECORATIVE_FAIL(ToAString(TX("Dimensions differ: ") + a + TX(" <=> ") + b).c_str());
							break;
						}
					case ErrLoadFailed:
						{
							DECORATIVE_FAIL(ToAString(TX("Failed to load: ") + a + TX(" <=> ") + b).c_str());
							break;
						}

					case ErrDataDiffer:
						{
							DECORATIVE_FAIL(ToAString(TX("Mismatch: ") + a + TX(" <=> ") + b).c_str());
							break;
						}
				}
			}
			catch(Err::Exception& e) {
				DECORATIVE_FAIL(ToAString(e.Desc()).c_str());

			}
		}
	}
}
