#include "StdAfx.h"
#include "orz/file.h"
#include "orz/fileops.h"
#include "main.h"

enum
{
	SeekDelta = 100
};

SUITE(IO)
{
	class ReaderTestFixture
	{
	public:
		ReaderTestFixture():file(g_datapath + TX("/data.raw"))
		{
			file.Open();
		}

		~ReaderTestFixture() {}

		IO::FileReader file;
	};

	class ReaderAsciiFixture
	{
	public:
		ReaderAsciiFixture()
			:content("This is a simple text file."), file(g_datapath + TX("/ascii.txt"))
		{
			file.Open();
		}

		~ReaderAsciiFixture() {}

		IO::FileReader file;
		AString content;
	};

	TEST(FileReader_EmptyState)
	{
		IO::FileReader f(g_datapath + TX("/ascii.txt"));
		CHECK(f.IsOpen() == false);	// Shouldn't be open
		CHECK_THROW(f.Position(), Err::FileNotOpen);	// Shouldn't be able to check position
		CHECK_THROW(f.Size(), Err::FileNotOpen); // Shouldn't be able to get file size

		Byte buf;
		CHECK_THROW(f.Read(&buf, 1, 1), Err::FileNotOpen);	// Read should fail
	}

	TEST(FileReader_Open)
	{
		IO::FileReader f(g_datapath + TX("/data.raw"));
		CHECK(f.Open());
	}

	TEST_FIXTURE(ReaderTestFixture, FileReader_OpenedState)
	{
		CHECK(file.IsOpen() == true);
		CHECK(file.Position() == 0);
	}

	TEST_FIXTURE(ReaderTestFixture, FileReader_SeekBegin)
	{
		file.Seek(SeekDelta, IO::SeekMethodBegin);
		CHECK(file.Position() == SeekDelta);
	}

	TEST_FIXTURE(ReaderTestFixture, FileReader_SeekCurrent)
	{
		// Should just move forward SeekDelta amount of bytes
		file.Seek(SeekDelta, IO::SeekMethodCurrent);
		CHECK(file.Position() == SeekDelta);

		// Should have moved a total of SeekDelta * 2 bytes
		file.Seek(SeekDelta, IO::SeekMethodCurrent);
		CHECK(file.Position() == SeekDelta * 2);
	}

	TEST_FIXTURE(ReaderTestFixture, FileReader_SeekEnd)
	{
		// Should be SeekDelta bytes remaining at the end
		file.Seek(-SeekDelta, IO::SeekMethodEnd);
		CHECK(file.Position() == static_cast<int>(file.Size() - SeekDelta));
	}

	TEST_FIXTURE(ReaderTestFixture, FileReader_ReadFullThrows)
	{
		file.Seek(-static_cast<long>(SeekDelta * 1.5), IO::SeekMethodEnd);
		Byte buffer[SeekDelta];

		// First one should work
		file.ReadFull(buffer, SeekDelta);

		// Second one shouldn't
		CHECK_THROW(file.ReadFull(buffer, SeekDelta), Err::EndOfFile);
	}

	TEST(FileReader_NoFile)
	{
		IO::FileReader f(TX("This file does not exist"));
		CHECK(f.Open() == false);
		//CHECK(f.Error() == IO::OpenFileNotFound);
		CHECK(f.IsOpen() == false);	// Shouldn't be open
	}

	TEST_FIXTURE(ReaderTestFixture, FileReader_OpenMany)
	{
		CHECK_THROW(file.Open(), Err::FileAlreadyOpen);
	}

	TEST(DoFileExist)
	{
		CHECK(IO::DoFileExist(g_datapath + TX("/data.raw")) == true);
		CHECK(IO::DoFileExist(TX("This file should not exist")) == false);
	}

	TEST_FIXTURE(ReaderAsciiFixture, ReadBytes)
	{
		AString txt;
		txt.reserve(27);
		AChar c;
		for(int i = 0; i < 27; ++i)
		{
			CHECK(file.Read(&c, 1, 1) == 1);
			txt.push_back(c);
		}

		CHECK(txt == "This is a simple text file.");
		CHECK(file.Read(&c, 1, 1) == 0);
	}

	TEST_FIXTURE(ReaderAsciiFixture, ReadByteSeek)
	{
		AString txt;
		txt.reserve(27);
		AChar c;
		for(int i = 0; i < 27; ++i)
		{
			CHECK(file.Read(&c, 1, 1) == 1);
			txt.push_back(c);
		}

		CHECK(txt == "This is a simple text file.");
		CHECK(file.Read(&c, 1, 1) == 0);
	}

	TEST_FIXTURE(ReaderAsciiFixture, ReadMuch)
	{
		AChar c[200];
		CHECK(file.Read(c, 1, 100) == 27);

		for(int i = 0; i < 27; ++i)
		{
			bool isEqual = (content.at(i) == c[i]);
			CHECK(isEqual);
			if (!isEqual) return;
		}

		CHECK(file.Read(c, 1, 1) == 0);
	}


	TEST_FIXTURE(ReaderTestFixture, ReadPosition)
	{
		Byte c;
		for (int i = 0; i < 12288; ++i)
		{
			bool couldRead = (file.Read(&c, 1, 1) == 1);
			bool correctPosition = (file.Position() == (i + 1));
			CHECK(couldRead);
			CHECK(correctPosition);
			if (couldRead == false || correctPosition == false) break;
		}
	}

	/*TEST(FileAccessViolation)
	{
		const Char* filename = TX("file_access_test.txt");
		IO::FileWriter writer;
		CHECK(writer.Open(filename) == true);

		IO::FileReader reader(filename);
		CHECK(reader.Open() == false);
		CHECK(reader.Error() == IO::OpenAccessViolation);
	}*/
}
